# -*- test-case-name: twisted.trial.test.test_log -*-

# fodder for test_script, which parses files for emacs local variable
# declarations.  This one is supposed to have:
#    test-case-name: twisted.trial.test.test_log.
# in the first line
# The class declaration is irrelevant


class Foo:
    pass
